# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Adapts the grpc surface."""

from __future__ import absolute_import
from grpc.beta import implementations
from grpc.framework.interfaces.face import face
from . import auth


API_ERRORS = (face.AbortionError, )
"""gRPC exceptions that indicate that an RPC was aborted."""


def _make_grpc_auth_func(auth_func):
    """Creates the auth func expected by the grpc callback."""

    def grpc_auth(dummy_context, callback):
        """The auth signature required by grpc."""
        callback(auth_func(), None)

    return grpc_auth


def _make_channel_creds(auth_func, ssl_creds):
    """Converts the auth func into the composite creds expected by grpc."""
    grpc_auth_func = _make_grpc_auth_func(auth_func)
    call_creds = implementations.metadata_call_credentials(grpc_auth_func)
    return implementations.composite_channel_credentials(ssl_creds, call_creds)


def create_stub(generated_create_stub, service_path, port, ssl_creds=None,
                channel=None, metadata_transformer=None, scopes=None):
    """Creates a gRPC client stub.

    Args:
        generated_create_stub: The generated gRPC method to create a stub.
        service_path: The domain name of the API remote host.
        port: The port on which to connect to the remote host.
        ssl_creds: A ClientCredentials object for use with an SSL-enabled
            Channel. If none, credentials are pulled from a default location.
        channel: A Channel object through which to make calls. If none, a secure
            channel is constructed.
        metadata_transformer: A function that transforms the metadata for
            requests, e.g., to give OAuth credentials.
        scopes: The OAuth scopes for this service. This parameter is ignored if
            a custom metadata_transformer is supplied.

    Returns:
        A gRPC client stub.
    """
    if channel is None:
        if ssl_creds is None:
            ssl_creds = implementations.ssl_channel_credentials(
                None, None, None)
        if metadata_transformer is None:
            if scopes is None:
                scopes = []
            metadata_transformer = auth.make_auth_func(scopes)

        channel_creds = _make_channel_creds(metadata_transformer, ssl_creds)
        channel = implementations.secure_channel(
            service_path, port, channel_creds)

    return generated_create_stub(channel)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides function wrappers that implement page streaming and retrying."""

from __future__ import absolute_import, division
import random
import sys
import time

from . import (BackoffSettings, BundleOptions, bundling, CallSettings, config,
               OPTION_INHERIT, RetryOptions)
from .errors import GaxError, RetryError

_MILLIS_PER_SECOND = 1000


def _add_timeout_arg(a_func, timeout):
    """Updates a_func so that it gets called with the timeout as its final arg.

    This converts a callable, a_func, into another callable with an additional
    positional arg.

    Args:
      a_func (callable): a callable to be updated
      timeout (int): to be added to the original callable as it final positional
        arg.


    Returns:
      callable: the original callable updated to the timeout arg
    """

    def inner(*args, **kw):
        """Updates args with the timeout."""
        updated_args = args + (timeout,)
        return a_func(*updated_args, **kw)

    return inner


def _retryable(a_func, retry):
    """Creates a function equivalent to a_func, but that retries on certain
    exceptions.

    Args:
        a_func (callable): A callable.
        retry (RetryOptions): Configures the exceptions upon which the callable
          should retry, and the parameters to the exponential backoff retry
          algorithm.

    Returns:
        A function that will retry on exception.
    """

    delay_mult = retry.backoff_settings.retry_delay_multiplier
    max_delay = (retry.backoff_settings.max_retry_delay_millis /
                 _MILLIS_PER_SECOND)
    timeout_mult = retry.backoff_settings.rpc_timeout_multiplier
    max_timeout = (retry.backoff_settings.max_rpc_timeout_millis /
                   _MILLIS_PER_SECOND)
    total_timeout = (retry.backoff_settings.total_timeout_millis /
                     _MILLIS_PER_SECOND)

    def inner(*args, **kwargs):
        """Equivalent to ``a_func``, but retries upon transient failure.

        Retrying is done through an exponential backoff algorithm configured
        by the options in ``retry``.
        """
        delay = retry.backoff_settings.initial_retry_delay_millis
        timeout = (retry.backoff_settings.initial_rpc_timeout_millis /
                   _MILLIS_PER_SECOND)
        exc = RetryError('Retry total timeout exceeded before any'
                         'response was received')
        now = time.time()
        deadline = now + total_timeout

        while now < deadline:
            try:
                to_call = _add_timeout_arg(a_func, timeout)
                return to_call(*args, **kwargs)

            # pylint: disable=broad-except
            except Exception as exception:
                if config.exc_to_code(exception) not in retry.retry_codes:
                    raise RetryError(
                        'Exception occurred in retry method that was not'
                        'classified as transient', exception)

                # pylint: disable=redefined-variable-type
                exc = RetryError('Retry total timeout exceeded with exception',
                                 exception)
                to_sleep = random.uniform(0, delay)
                time.sleep(to_sleep / _MILLIS_PER_SECOND)
                now = time.time()
                delay = min(delay * delay_mult, max_delay)
                timeout = min(
                    timeout * timeout_mult, max_timeout, deadline - now)
                continue

        raise exc

    return inner


def _bundleable(a_func, desc, bundler):
    """Creates a function that transforms an API call into a bundling call.

    It transform a_func from an API call that receives the requests and returns
    the response into a callable that receives the same request, and
    returns a :class:`bundling.Event`.

    The returned Event object can be used to obtain the eventual result of the
    bundled call.

    Args:
        a_func (callable[[req], resp]): an API call that supports bundling.
        desc (gax.BundleDescriptor): describes the bundling that a_func
          supports.
        bundler (gax.bundling.Executor): orchestrates bundling.

    Returns:
        callable: takes the API call's request and keyword args and returns a
          bundling.Event object.

    """
    def inner(*args, **kwargs):
        """Schedules execution of a bundling task."""
        request = args[0]
        the_id = bundling.compute_bundle_id(
            request, desc.request_discriminator_fields)
        return bundler.schedule(a_func, the_id, desc, request, kwargs)

    return inner


def _page_streamable(a_func,
                     request_page_token_field,
                     response_page_token_field,
                     resource_field):
    """Creates a function that yields an iterable to performs page-streaming.

    Args:
        a_func: an API call that is page streaming.
        request_page_token_field: The field of the page token in the request.
        response_page_token_field: The field of the next page token in the
          response.
        resource_field: The field to be streamed.

    Returns:
        A function that returns an iterable over the specified field.
    """

    def inner(*args, **kwargs):
        """A generator that yields all the paged responses."""
        request = args[0]
        while True:
            response = a_func(request, **kwargs)
            for obj in getattr(response, resource_field):
                yield obj
            next_page_token = getattr(response, response_page_token_field)
            if not next_page_token:
                break
            setattr(request, request_page_token_field, next_page_token)

    return inner


def _construct_bundling(method_config, method_bundling_override,
                        bundle_descriptor):
    """Helper for ``construct_settings()``.

    Args:
      method_config: A dictionary representing a single ``methods`` entry of the
        standard API client config file. (See ``construct_settings()`` for
        information on this yaml.)
      method_retry_override: A BundleOptions object, OPTION_INHERIT, or None.
        If set to OPTION_INHERIT, the retry settings are derived from method
        config. Otherwise, this parameter overrides ``method_config``.
      bundle_descriptor: A BundleDescriptor object describing the structure of
        bundling for this method. If not set, this method will not bundle.

    Returns:
      A tuple (bundling.Executor, BundleDescriptor) that configures bundling.
      The bundling.Executor may be None if this method should not bundle.
    """
    if 'bundling' in method_config and bundle_descriptor:

        if method_bundling_override == OPTION_INHERIT:
            bundler = bundling.Executor(BundleOptions(
                **method_config['bundling']))
        elif method_bundling_override:
            bundler = bundling.Executor(method_bundling_override)
        else:
            bundler = None

    else:
        bundler = None

    return bundler


def _construct_retry(
        method_config, method_retry_override, retry_codes, retry_params,
        retry_names):
    """Helper for ``construct_settings()``.

    Args:
      method_config: A dictionary representing a single ``methods`` entry of the
        standard API client config file. (See ``construct_settings()`` for
        information on this yaml.)
      method_retry_override: A RetryOptions object, OPTION_INHERIT, or None.
        If set to OPTION_INHERIT, the retry settings are derived from method
        config. Otherwise, this parameter overrides ``method_config``.
      retry_codes_def: A dictionary parsed from the ``retry_codes_def`` entry
        of the standard API client config file. (See ``construct_settings()``
        for information on this yaml.)
      retry_params: A dictionary parsed from the ``retry_params`` entry
        of the standard API client config file. (See ``construct_settings()``
        for information on this yaml.)
      retry_names: A dictionary mapping the string names used in the
        standard API client config file to API response status codes.

    Returns:
      A RetryOptions object, or None.
    """
    if method_retry_override != OPTION_INHERIT:
        return method_retry_override

    codes = []
    if retry_codes:
        for codes_name in retry_codes:
            if (codes_name == method_config['retry_codes_name'] and
                    retry_codes[codes_name]):
                codes = [
                    retry_names[name] for name in retry_codes[codes_name]]
                break

    params_struct = None
    if method_config.get('retry_params_name'):
        for params_name in retry_params:
            if params_name == method_config['retry_params_name']:
                params_struct = retry_params[params_name].copy()
                break
        backoff_settings = BackoffSettings(**params_struct)
    else:
        backoff_settings = None

    retry = RetryOptions(retry_codes=codes, backoff_settings=backoff_settings)
    return retry


def construct_settings(
        service_name, client_config, bundling_override, retry_override,
        retry_names, timeout, bundle_descriptors=None, page_descriptors=None):
    """Constructs a dictionary mapping method names to CallSettings.

    The ``client_config`` parameter is parsed from a client configuration JSON
    file of the form:

    .. code-block:: json

       {
         "interfaces": {
           "google.fake.v1.ServiceName": {
             "retry_codes": {
               "idempotent": ["UNAVAILABLE", "DEADLINE_EXCEEDED"],
               "non_idempotent": []
             },
             "retry_params": {
               "default": {
                 "initial_retry_delay_millis": 100,
                 "retry_delay_multiplier": 1.2,
                 "max_retry_delay_millis": 1000,
                 "initial_rpc_timeout_millis": 2000,
                 "rpc_timeout_multiplier": 1.5,
                 "max_rpc_timeout_millis": 30000,
                 "total_timeout_millis": 45000
               }
             },
             "methods": {
               "CreateFoo": {
                 "retry_codes_name": "idempotent",
                 "retry_params_name": "default"
               },
               "Publish": {
                 "retry_codes_name": "non_idempotent",
                 "retry_params_name": "default",
                 "bundling": {
                   "element_count_threshold": 40,
                   "element_count_limit": 200,
                   "request_byte_threshold": 90000,
                   "request_byte_limit": 100000,
                   "delay_threshold_millis": 100
                 }
               }
             }
           }
         }
       }

    Args:
      service_name: The fully-qualified name of this service, used as a key into
       the client config file (in the example above, this value should be
       ``google.fake.v1.ServiceName``).
      client_config: A dictionary parsed from the standard API client config
       file.
      bundle_descriptors: A dictionary of method names to BundleDescriptor
       objects for methods that are bundling-enabled.
      page_descriptors: A dictionary of method names to PageDescriptor objects
       for methods that are page streaming-enabled.
      bundling_override: A dictionary of method names to BundleOptions
        override those specified in ``client_config``.
      retry_override: A dictionary of method names to RetryOptions that
        override those specified in ``client_config``.
      retry_names: A dictionary mapping the strings referring to response status
        codes to the Python objects representing those codes.
      timeout: The timeout parameter for all API calls in this dictionary.

    Raises:
      KeyError: If the configuration for the service in question cannot be
        located in the provided ``client_config``.
    """
    defaults = dict()
    bundle_descriptors = bundle_descriptors or {}
    page_descriptors = page_descriptors or {}

    try:
        service_config = client_config['interfaces'][service_name]
    except KeyError:
        raise KeyError('Client configuration not found for service: {}'
                       .format(service_name))

    for method in service_config.get('methods'):
        method_config = service_config['methods'][method]

        bundle_descriptor = bundle_descriptors.get(method)
        bundler = _construct_bundling(
            method_config, bundling_override.get(method, OPTION_INHERIT),
            bundle_descriptor)

        retry = _construct_retry(
            method_config, retry_override.get(method, OPTION_INHERIT),
            service_config['retry_codes'], service_config['retry_params'],
            retry_names)

        defaults[method] = CallSettings(
            timeout=timeout, retry=retry,
            page_descriptor=page_descriptors.get(method),
            bundler=bundler, bundle_descriptor=bundle_descriptor)

    return defaults


def _catch_errors(a_func, errors):
    """Updates a_func to wrap exceptions with GaxError

    Args:
        a_func (callable): A callable.
        retry (list[Exception]): Configures the exceptions to wrap.

    Returns:
        A function that will wrap certain exceptions with GaxError
    """
    def inner(*args, **kwargs):
        """Wraps specified exceptions"""
        try:
            return a_func(*args, **kwargs)
        # pylint: disable=catching-non-exception
        except tuple(errors) as exception:
            raise (GaxError('RPC failed', cause=exception), None,
                   sys.exc_info()[2])

    return inner


def create_api_call(func, settings):
    """Converts an rpc call into an API call governed by the settings.

    In typical usage, ``func`` will be a callable used to make an rpc request.
    This will mostly likely be a bound method from a request stub used to make
    an rpc call.

    The result is created by applying a series of function decorators defined
    in this module to ``func``.  ``settings`` is used to determine which
    function decorators to apply.

    The result is another callable which for most values of ``settings`` has
    has the same signature as the original. Only when ``settings`` configures
    bundling does the signature change.

    Args:
      func (callable[[object], object]): is used to make a bare rpc call
      settings (:class:`CallSettings`): provides the settings for this call

    Returns:
      func (callable[[object], object]): a bound method on a request stub used
        to make an rpc call

    Raises:
       ValueError: if ``settings`` has incompatible values, e.g, if bundling
         and page_streaming are both configured

    """
    if settings.retry and settings.retry.retry_codes:
        api_call = _retryable(func, settings.retry)
    else:
        api_call = _add_timeout_arg(func, settings.timeout)

    if settings.page_descriptor:
        if settings.bundler and settings.bundle_descriptor:
            raise ValueError('ApiCallable has incompatible settings: '
                             'bundling and page streaming')
        return _page_streamable(
            api_call,
            settings.page_descriptor.request_page_token_field,
            settings.page_descriptor.response_page_token_field,
            settings.page_descriptor.resource_field)

    if settings.bundler and settings.bundle_descriptor:
        return _bundleable(api_call, settings.bundle_descriptor,
                           settings.bundler)

    return _catch_errors(api_call, config.API_ERRORS)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for gax package globals."""

from __future__ import absolute_import

import unittest2

from google.gax import (
    BundleOptions, CallOptions, CallSettings, OPTION_INHERIT, RetryOptions)


class TestBundleOptions(unittest2.TestCase):

    def test_cannont_construct_with_noarg_options(self):
        self.assertRaises(AssertionError,
                          BundleOptions)

    def test_cannont_construct_with_bad_options(self):
        not_an_int = 'i am a string'
        self.assertRaises(AssertionError,
                          BundleOptions,
                          element_count_threshold=not_an_int)
        self.assertRaises(AssertionError,
                          BundleOptions,
                          request_byte_threshold=not_an_int)
        self.assertRaises(AssertionError,
                          BundleOptions,
                          delay_threshold=not_an_int)


class TestCallSettings(unittest2.TestCase):

    def test_call_options_simple(self):
        options = CallOptions(timeout=23)
        self.assertEqual(options.timeout, 23)
        self.assertEqual(options.retry, OPTION_INHERIT)
        self.assertEqual(options.is_page_streaming, OPTION_INHERIT)

    def test_settings_merge_options1(self):
        retry = RetryOptions(None, None)
        options = CallOptions(timeout=46, retry=retry)
        settings = CallSettings(timeout=9, page_descriptor=None, retry=None)
        final = settings.merge(options)
        self.assertEqual(final.timeout, 46)
        self.assertEqual(final.retry, retry)
        self.assertIsNone(final.page_descriptor)

    def test_settings_merge_options2(self):
        options = CallOptions(retry=None)
        settings = CallSettings(
            timeout=9, page_descriptor=None, retry=RetryOptions(None, None))
        final = settings.merge(options)
        self.assertEqual(final.timeout, 9)
        self.assertIsNone(final.page_descriptor)
        self.assertIsNone(final.retry)

    def test_settings_merge_options_page_streaming(self):
        retry = RetryOptions(None, None)
        options = CallOptions(timeout=46, is_page_streaming=False)
        settings = CallSettings(timeout=9, retry=retry)
        final = settings.merge(options)
        self.assertEqual(final.timeout, 46)
        self.assertIsNone(final.page_descriptor)
        self.assertEqual(final.retry, retry)

    def test_settings_merge_none(self):
        settings = CallSettings(
            timeout=23, page_descriptor=object(), bundler=object(),
            retry=object())
        final = settings.merge(None)
        self.assertEqual(final.timeout, settings.timeout)
        self.assertEqual(final.retry, settings.retry)
        self.assertEqual(final.page_descriptor, settings.page_descriptor)
        self.assertEqual(final.bundler, settings.bundler)
        self.assertEqual(final.bundle_descriptor, settings.bundle_descriptor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import sys
from setuptools import setup, find_packages

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

# Get the version
version_regex = r'__version__ = ["\']([^"\']*)["\']'
with open('google/gax/__init__.py', 'r') as f:
    text = f.read()
    match = re.search(version_regex, text)
    if match:
        version = match.group(1)
    else:
        raise RuntimeError("No version number found!")

install_requires = [
    'grpcio==0.13.0',
    'ply==3.8',
    'protobuf>=3.0.0b1.post1',
    'oauth2client>=1.5.2',
]

setup(
    name='google-gax',
    version=version,
    description='Google API Extensions',
    long_description=open('README.rst').read(),
    author='Google API Authors',
    author_email='googleapis-packages@google.com',
    url='https://github.com/googleapis/gax-python',
    namespace_packages = ['google'],
    packages=find_packages(),
    package_dir={'google-gax': 'google'},
    license='BSD-3-Clause',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    tests_require=['pytest'],
    install_requires=install_requires,
)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Google API Extensions"""

from __future__ import absolute_import
import collections


__version__ = '0.9.0'


OPTION_INHERIT = object()
"""Global constant.

If a CallOptions field is set to OPTION_INHERIT, the call to which that
CallOptions belongs will attempt to inherit that field from its default
settings."""


class CallSettings(object):
    """Encapsulates the call settings for an ApiCallable"""
    # pylint: disable=too-few-public-methods
    def __init__(self, timeout=30, retry=None, page_descriptor=None,
                 bundler=None, bundle_descriptor=None):
        """Constructor.

        Args:
            timeout (int): The client-side timeout for API calls. This
              parameter is ignored for retrying calls.
            retry (RetryOptions): The configuration for retrying upon transient
              error. If set to None, this call will not retry.
            page_descriptor (PageDescriptor): indicates the structure of page
              streaming to be performed. If set to None, page streaming is
              not performed.
            bundler (bundle.Executor): orchestrates bundling. If None, bundling
              is not performed.
            bundle_descriptor (BundleDescriptor): indicates the structure of
              the bundle. If None, bundling is not performed.

        Returns:
            A CallSettings object.
        """
        self.timeout = timeout
        self.retry = retry
        self.page_descriptor = page_descriptor
        self.bundler = bundler
        self.bundle_descriptor = bundle_descriptor

    def merge(self, options):
        """Returns a new CallSettings merged from this and a CallOptions object.

        Args:
            options: A CallOptions object whose values are override those in
              this object. If None, `merge` returns a copy of this object.

        Returns:
            A CallSettings object.
        """
        if not options:
            return CallSettings(
                timeout=self.timeout, retry=self.retry,
                page_descriptor=self.page_descriptor, bundler=self.bundler,
                bundle_descriptor=self.bundle_descriptor)
        else:
            if options.timeout == OPTION_INHERIT:
                timeout = self.timeout
            else:
                timeout = options.timeout

            if options.retry == OPTION_INHERIT:
                retry = self.retry
            else:
                retry = options.retry

            if options.is_page_streaming:
                page_descriptor = self.page_descriptor
            else:
                page_descriptor = None

            return CallSettings(
                timeout=timeout, retry=retry,
                page_descriptor=page_descriptor, bundler=self.bundler,
                bundle_descriptor=self.bundle_descriptor)


class CallOptions(object):
    """Encapsulates the overridable settings for a particular API call"""
    # pylint: disable=too-few-public-methods
    def __init__(self, timeout=OPTION_INHERIT, retry=OPTION_INHERIT,
                 is_page_streaming=OPTION_INHERIT):
        """Constructor.

        Args:
            timeout (int): The client-side timeout for API calls.
            retry (RetryOptions): The configuration for retrying upon transient error.
              If set to None, this call will not retry.
            is_page_streaming (bool): If set and the call is configured for page
              streaming, page streaming is performed.

        Returns:
            A CallOptions object.
        """
        self.timeout = timeout
        self.retry = retry
        self.is_page_streaming = is_page_streaming


class PageDescriptor(
        collections.namedtuple(
            'PageDescriptor',
            ['request_page_token_field',
             'response_page_token_field',
             'resource_field'])):
    """Describes the structure of a page-streaming call."""
    pass


class RetryOptions(
        collections.namedtuple(
            'RetryOptions',
            ['retry_codes',
             'backoff_settings'])):
    """Per-call configurable settings for retrying upon transient failure.

    Attributes:
      retry_codes: a list of exceptions upon which a retry should be attempted.
      backoff_settings: a BackoffSettings object configuring the retry
        exponential backoff algorithm.
    """
    pass


class BackoffSettings(
        collections.namedtuple(
            'BackoffSettings',
            ['initial_retry_delay_millis',
             'retry_delay_multiplier',
             'max_retry_delay_millis',
             'initial_rpc_timeout_millis',
             'rpc_timeout_multiplier',
             'max_rpc_timeout_millis',
             'total_timeout_millis'])):
    """Parameters to the exponential backoff algorithm for retrying.

    Attributes:
      initial_retry_delay_millis: the initial delay time, in milliseconds,
        between the completion of the first failed request and the initiation of
        the first retrying request.
      retry_delay_multiplier: the multiplier by which to increase the delay time
        between the completion of failed requests, and the initiation of the
        subsequent retrying request.
      max_retry_delay_millis: the maximum delay time, in milliseconds, between
        requests. When this value is reached, ``retry_delay_multiplier`` will no
        longer be used to increase delay time.
      initial_rpc_timeout_millis: the initial timeout parameter to the request.
      rpc_timeout_multiplier: the multiplier by which to increase the timeout
        parameter between failed requests.
      max_rpc_timeout_millis: the maximum timeout parameter, in milliseconds,
        for a request. When this value is reached, ``rpc_timeout_multiplier``
        will no longer be used to increase the timeout.
      total_timeout_millis: the total time, in milliseconds, starting from when
        the initial request is sent, after which an error will be returned,
        regardless of the retrying attempts made meanwhile.
    """
    pass


class BundleDescriptor(
        collections.namedtuple(
            'BundleDescriptor',
            ['bundled_field',
             'request_discriminator_fields',
             'subresponse_field'])):
    """Describes the structure of bundled call.

    request_discriminator_fields may include '.' as a separator, which is used
    to indicate object traversal.  This allows fields in nested objects to be
    used to determine what requests to bundle.

    Attributes:
      bundled_field: the repeated field in the request message that
        will have its elements aggregated by bundling
      request_discriminator_fields: a list of fields in the
        target request message class that are used to determine
        which messages should be bundled together.
      subresponse_field: an optional field, when present it indicates the field
        in the response message that should be used to demultiplex the response
        into multiple response messages.
    """
    def __new__(cls,
                bundled_field,
                request_discriminator_fields,
                subresponse_field=None):
        return super(cls, BundleDescriptor).__new__(
            cls,
            bundled_field,
            request_discriminator_fields,
            subresponse_field)


class BundleOptions(
        collections.namedtuple(
            'BundleOptions',
            ['element_count_threshold',
             'element_count_limit',
             'request_byte_threshold',
             'request_byte_limit',
             'delay_threshold'])):
    """Holds values used to configure bundling.

    The xxx_threshold attributes are used to configure when the bundled request
    should be made.

    Attributes:
        element_count_threshold: the bundled request will be sent once the
          count of outstanding elements in the repeated field reaches this
          value.
        element_count_limit: represents a hard limit on the number of elements
          in the repeated field of the bundle; if adding a request to a bundle
          would exceed this value, the bundle is sent and the new request is
          added to a fresh bundle. It is invalid for a single request to exceed
          this limit.
        request_byte_threshold: the bundled request will be sent once the count
          of bytes in the request reaches this value. Note that this value is
          pessimistically approximated by summing the bytesizes of the elements
          in the repeated field, and therefore may be an under-approximation.
        request_byte_limit: represents a hard limit on the size of the bundled
          request; if adding a request to a bundle would exceed this value, the
          bundle is sent and the new request is added to a fresh bundle. It is
          invalid for a single request to exceed this limit. Note that this
          value is pessimistically approximated by summing the bytesizes of the
          elements in the repeated field, with a buffer applied to correspond to
          the resulting under-approximation.
        delay_threshold: the bundled request will be sent this amount of
          time after the first element in the bundle was added to it.

    """
    # pylint: disable=too-few-public-methods

    def __new__(cls,
                element_count_threshold=0,
                element_count_limit=0,
                request_byte_threshold=0,
                request_byte_limit=0,
                delay_threshold=0):
        """Invokes the base constructor with default values.

        The default values are zero for all attributes and it's necessary to
        specify at least one valid threshold value during construction.

        Args:
           element_count_threshold: the bundled request will be sent once the
             count of outstanding elements in the repeated field reaches this
             value.
           element_count_limit: represents a hard limit on the number of
             elements in the repeated field of the bundle; if adding a request
             to a bundle would exceed this value, the bundle is sent and the new
             request is added to a fresh bundle. It is invalid for a single
             request to exceed this limit.
           request_byte_threshold: the bundled request will be sent once the
             count of bytes in the request reaches this value. Note that this
             value is pessimistically approximated by summing the bytesizes of
             the elements in the repeated field, with a buffer applied to
             compensate for the corresponding under-approximation.
           request_byte_limit: represents a hard limit on the size of the
             bundled request; if adding a request to a bundle would exceed this
             value, the bundle is sent and the new request is added to a fresh
             bundle. It is invalid for a single request to exceed this
             limit. Note that this value is pessimistically approximated by
             summing the bytesizes of the elements in the repeated field, with a
             buffer applied to correspond to the resulting under-approximation.
           delay_threshold: the bundled request will be sent this amount of
             time after the first element in the bundle was added to it.

        """
        assert isinstance(element_count_threshold, int), 'should be an int'
        assert isinstance(element_count_limit, int), 'should be an int'
        assert isinstance(request_byte_threshold, int), 'should be an int'
        assert isinstance(request_byte_limit, int), 'should be an int'
        assert isinstance(delay_threshold, int), 'should be an int'
        assert (element_count_threshold > 0 or
                request_byte_threshold > 0 or
                delay_threshold > 0), 'one threshold should be > 0'

        return super(cls, BundleOptions).__new__(
            cls,
            element_count_threshold,
            element_count_limit,
            request_byte_threshold,
            request_byte_limit,
            delay_threshold)



# Copyright 2016, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for grpc."""

from __future__ import absolute_import

import mock
import unittest2

from google.gax import grpc


def _fake_create_stub(channel):
    return channel


class TestCreateStub(unittest2.TestCase):
    FAKE_SERVICE_PATH = 'service_path'
    FAKE_PORT = 10101

    @mock.patch('grpc.beta.implementations.composite_channel_credentials')
    @mock.patch('grpc.beta.implementations.ssl_channel_credentials')
    @mock.patch('grpc.beta.implementations.secure_channel')
    @mock.patch('google.gax.auth.make_auth_func')
    def test_creates_a_stub_ok_with_no_scopes(
            self, auth, chan, chan_creds, comp):
        got_channel = grpc.create_stub(
            _fake_create_stub, self.FAKE_SERVICE_PATH, self.FAKE_PORT)
        chan_creds.assert_called_once_with(None, None, None)
        chan.assert_called_once_with(self.FAKE_SERVICE_PATH, self.FAKE_PORT,
                                     comp.return_value)
        auth.assert_called_once_with([])
        self.assertEquals(got_channel, chan.return_value)

    @mock.patch('grpc.beta.implementations.composite_channel_credentials')
    @mock.patch('grpc.beta.implementations.ssl_channel_credentials')
    @mock.patch('grpc.beta.implementations.secure_channel')
    @mock.patch('google.gax.auth.make_auth_func')
    def test_creates_a_stub_ok_with_scopes(
            self, auth, chan, chan_creds, comp):
        fake_scopes = ['dummy', 'scopes']
        grpc.create_stub(
            _fake_create_stub, self.FAKE_SERVICE_PATH, self.FAKE_PORT,
            scopes=fake_scopes)
        chan_creds.assert_called_once_with(None, None, None)
        chan.assert_called_once_with(self.FAKE_SERVICE_PATH, self.FAKE_PORT,
                                     comp.return_value)
        auth.assert_called_once_with(fake_scopes)

    @mock.patch('grpc.beta.implementations.metadata_call_credentials')
    @mock.patch('grpc.beta.implementations.composite_channel_credentials')
    @mock.patch('grpc.beta.implementations.ssl_channel_credentials')
    @mock.patch('grpc.beta.implementations.secure_channel')
    @mock.patch('google.gax.auth.make_auth_func')
    def test_creates_a_stub_with_given_channel(
            self, auth, chan, chan_creds, comp, md):
        fake_channel = object()
        got_channel = grpc.create_stub(
            _fake_create_stub, self.FAKE_SERVICE_PATH, self.FAKE_PORT,
            channel=fake_channel)
        self.assertEquals(got_channel, fake_channel)
        self.assertFalse(auth.called)
        self.assertFalse(chan_creds.called)
        self.assertFalse(chan.called)
        self.assertFalse(comp.called)
        self.assertFalse(md.called)

    @mock.patch('grpc.beta.implementations.metadata_call_credentials')
    @mock.patch('grpc.beta.implementations.composite_channel_credentials')
    @mock.patch('grpc.beta.implementations.ssl_channel_credentials')
    @mock.patch('grpc.beta.implementations.secure_channel')
    @mock.patch('google.gax.auth.make_auth_func')
    def test_creates_a_stub_ok_with_given_creds(self, auth, chan, chan_creds,
                                                comp, md):
        fake_creds = object()
        got_channel = grpc.create_stub(
            _fake_create_stub, self.FAKE_SERVICE_PATH, self.FAKE_PORT,
            ssl_creds=fake_creds)
        chan.assert_called_once_with(self.FAKE_SERVICE_PATH, self.FAKE_PORT,
                                     comp.return_value)
        auth.assert_called_once_with([])
        self.assertTrue(chan.called)
        self.assertFalse(chan_creds.called)
        self.assertTrue(comp.called)
        self.assertTrue(md.called)
        self.assertEquals(got_channel, chan.return_value)

    @mock.patch('grpc.beta.implementations.composite_channel_credentials')
    @mock.patch('grpc.beta.implementations.ssl_channel_credentials')
    @mock.patch('grpc.beta.implementations.secure_channel')
    @mock.patch('google.gax.auth.make_auth_func')
    def test_creates_a_stub_ok_with_given_auth_func(self, auth, dummy_chan,
                                                    dummy_chan_creds, dummy_md):
        grpc.create_stub(
            _fake_create_stub, self.FAKE_SERVICE_PATH, self.FAKE_PORT,
            metadata_transformer=lambda x: tuple())
        self.assertFalse(auth.called)
